local M = {}

M.enabled = false
M.decreaseToInitial = false
M.decrease = false
M.speed = 0.001
M.limit = nil
M.ocean = nil

local initialWaterPosition = nil
local oldMission = nil

-- functions
local function getObject(className, preferredObjName)
    local objNames = scenetree.findClassObjects(className)
    if objNames and tableSize(objNames) > 0 then
        for _, name in pairs(objNames) do
            local obj = scenetree.findObject(name)
            if obj and (name == preferredObjName or not preferredObjName) then
                return obj
            end
        end
    else
        local obj = scenetree.findObject(className)
        if obj then return obj end
    end

    return nil
end

local function matrixToTable(mat)
    return {
        c0 = mat:getColumn(0),
        c1 = mat:getColumn(1),
        c2 = mat:getColumn(2),
        c3 = mat:getColumn(3)
    }
end

local function tableToMatrix(tbl)
    local mat = MatrixF(true)
    mat:setColumn(0, tbl.c0)
    mat:setColumn(1, tbl.c1)
    mat:setColumn(2, tbl.c2)
    mat:setColumn(3, tbl.c3)
    return mat
end

local function setup()
    M.enabled = false
    M.speed = 0.001
    M.limit = nil
    M.decrease = false

    M.ocean = nil
    initialWaterPosition = nil
end

local function getOcean()
    local waterPlane = getObject("WaterPlane")
    if waterPlane then
        initialWaterPosition = waterPlane.position
    end

    return waterPlane
end

local function resetWaterLevel()
    if not M.ocean then return end
    M.ocean.position = initialWaterPosition
end

local function getWaterLevel()
    if not M.ocean then return end
    return M.ocean.position:getColumn(3).z
end

local function setWaterLevel(level)
    if not M.ocean then return end
    local c3 = M.ocean.position:getColumn(3)
    M.ocean.position = tableToMatrix({
        c0 = M.ocean.position:getColumn(0),
        c1 = M.ocean.position:getColumn(1),
        c2 = M.ocean.position:getColumn(2),
        c3 = vec3(c3.x, c3.y, level)
    })
end

-- callbacks
local function onUpdate()
    local mission = getMissionFilename()
    if not oldMission or oldMission ~= mission then
        if mission == "" then return end
        oldMission = mission

        setup()
        M.ocean = getOcean()
    end

    if not M.enabled or not M.ocean then return end

    local height = getWaterLevel()

    if M.decreaseToInitial then
        local initialZ = string.format("%.2f", initialWaterPosition:getColumn(3).z)
        local z = string.format("%.2f", height)

        if z > initialZ then
            height = height - M.speed
        elseif z < initialZ then
            height = height + M.speed
        else
            M.enabled = false
            M.decreaseToInitial = false
        end
    else
        if M.decrease then
            local target = height - M.speed
            
            if M.limit then
                if target > M.limit then
                    height = target
                end
            else
                height = height - M.speed
            end
        else
            local target = height + M.speed
            
            if M.limit then
                if target < M.limit then
                    height = target
                end
            else
                height = height + M.speed
            end
        end
    end

    setWaterLevel(height)
end

local function onExtensionUnloaded()
    resetWaterLevel()
    setup()
end

M.resetWaterLevel = resetWaterLevel
M.getWaterLevel = getWaterLevel
M.setWaterLevel = setWaterLevel

M.onExtensionUnloaded = onExtensionUnloaded
M.onUpdate = onUpdate

return M